package pe

import (
	"encoding/binary"
	"io"

	"git.akyoto.dev/cli/q/src/config"
	"git.akyoto.dev/cli/q/src/os/common"
)

// EXE is the portable executable format used on Windows.
type EXE struct {
	DOSHeader
	NTHeader
	OptionalHeader64
	CodeHeader SectionHeader
	Code       []byte
	Data       []byte
}

// New creates a new EXE file.
func New(code []byte, data []byte) *EXE {
	const codeStart = 0x170
	const optHeaderSize = 0xF0

	codeSize := uint32(len(code))
	headerSize := uint32(codeStart)
	sectionAlign := uint32(0x10)
	fileAlign := uint32(0x10)
	imageSize := uint32(codeStart + len(code))
	imageSize += common.Padding(imageSize, sectionAlign)

	return &EXE{
		DOSHeader: DOSHeader{
			Magic:          [4]byte{'M', 'Z', 0, 0},
			NTHeaderOffset: 0x40,
		},
		NTHeader: NTHeader{
			Signature:            [4]byte{'P', 'E', 0, 0},
			Machine:              IMAGE_FILE_MACHINE_AMD64,
			NumberOfSections:     1,
			SizeOfOptionalHeader: optHeaderSize,
			Characteristics:      IMAGE_FILE_EXECUTABLE_IMAGE | IMAGE_FILE_LARGE_ADDRESS_AWARE,
		},
		OptionalHeader64: OptionalHeader64{
			Magic:                       0x020B, // PE32+ executable
			MajorLinkerVersion:          0x0E,
			MinorLinkerVersion:          0x16,
			SizeOfCode:                  codeSize,
			AddressOfEntryPoint:         codeStart,
			ImageBase:                   config.BaseAddress,
			SectionAlignment:            sectionAlign, // power of 2, must be greater than or equal to FileAlignment
			FileAlignment:               fileAlign,    // power of 2
			MajorOperatingSystemVersion: 0x06,
			MajorSubsystemVersion:       0x06,
			SizeOfImage:                 imageSize,
			SizeOfHeaders:               headerSize,
			Subsystem:                   IMAGE_SUBSYSTEM_WINDOWS_GUI,
			DllCharacteristics:          IMAGE_DLLCHARACTERISTICS_HIGH_ENTROPY_VA | IMAGE_DLLCHARACTERISTICS_DYNAMIC_BASE | IMAGE_DLLCHARACTERISTICS_NX_COMPAT | IMAGE_DLLCHARACTERISTICS_TERMINAL_SERVER_AWARE,
			SizeOfStackReserve:          0x100000,
			SizeOfStackCommit:           0x1000,
			SizeOfHeapReserve:           0x100000,
			SizeOfHeapCommit:            0x1000,
			NumberOfRvaAndSizes:         16,
			DataDirectory: [16]DataDirectory{
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
				{VirtualAddress: 0, Size: 0},
			},
		},
		CodeHeader: SectionHeader{
			Name:            [8]byte{'.', 't', 'e', 'x', 't'},
			VirtualSize:     uint32(len(code)),
			VirtualAddress:  codeStart,
			RawSize:         uint32(len(code)), // must be a multiple of FileAlignment
			RawAddress:      codeStart,         // must be a multiple of FileAlignment
			Characteristics: IMAGE_SCN_CNT_CODE | IMAGE_SCN_MEM_EXECUTE | IMAGE_SCN_MEM_READ,
		},
		Code: code,
		// Data: data,
	}
}

// Write writes the EXE file to the given writer.
func (pe *EXE) Write(writer io.Writer) {
	binary.Write(writer, binary.LittleEndian, &pe.DOSHeader)
	binary.Write(writer, binary.LittleEndian, &pe.NTHeader)
	binary.Write(writer, binary.LittleEndian, &pe.OptionalHeader64)
	binary.Write(writer, binary.LittleEndian, &pe.CodeHeader)
	binary.Write(writer, binary.LittleEndian, &pe.Code)
	// binary.Write(writer, binary.LittleEndian, &pe.Data)
}
